"""Temporary helper functions to convert from XML to a YAML structure."""

from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Tuple,
    Union,
)


def to_dynamics(recipe: Dict[str, Any]) -> Dict[str, Any]:
    def determine_type(definition: Dict[str, Any]) -> str:
        decay_type = definition["Type"]
        decay_type_xml_to_yaml = {
            "relativisticBreitWigner": "RelativisticBreitWigner",
            "nonResonant": "NonDynamic",
        }
        for xml_type, yaml_type in decay_type_xml_to_yaml.items():
            decay_type = decay_type.replace(xml_type, yaml_type)
        return decay_type

    def determine_form_factor(
        definition: Dict[str, Any],
        particle_xml: Dict[str, Any],
        decay_type: str,
    ) -> Dict[str, Any]:
        form_factor = definition.get("FormFactor", None)
        if form_factor is None:
            if decay_type == "NonDynamic":
                form_factor = {
                    "Type": "BlattWeisskopf",
                    "MesonRadius": 1.0,
                }
        else:
            parameters = _safe_wrap_in_list(
                particle_xml["DecayInfo"]["Parameter"]
            )
            meson_radius_candidates = [
                par for par in parameters if par["Type"] == "MesonRadius"
            ]
            if meson_radius_candidates:
                meson_radius_xml = meson_radius_candidates[0]
                meson_radius_value = float(meson_radius_xml["Value"])
                if len(meson_radius_xml) == 1:
                    form_factor["MesonRadius"] = meson_radius_value
                else:
                    meson_radius_yml = {"Value": meson_radius_value}
                    optional_keys = [
                        ("Min", float),
                        ("Max", float),
                        ("Fix", bool),
                    ]
                    for key, converter in optional_keys:
                        if key in meson_radius_xml:
                            meson_radius_yml[key] = converter(
                                meson_radius_xml[key]
                            )
                    form_factor["MesonRadius"] = meson_radius_yml
        return form_factor

    particle_list_xml = recipe["ParticleList"]["Particle"]
    dynamics_yml: Dict[str, Any] = dict()
    for particle_xml in particle_list_xml:
        name = particle_xml["Name"]
        decay_info_xml = particle_xml.get("DecayInfo", None)
        if decay_info_xml:
            decay_type = determine_type(decay_info_xml)
            form_factor = determine_form_factor(
                decay_info_xml, particle_xml, decay_type
            )
            decay_info_yml = {"Type": decay_type, "FormFactor": form_factor}
            dynamics_yml[name] = decay_info_yml
    return dynamics_yml


def to_intensity(recipe: Dict[str, Any]) -> Dict[str, Any]:
    intensity_xml = recipe["Intensity"]
    intensity_yml = _extract_intensity_component(intensity_xml)
    return intensity_yml


def to_kinematics_dict(recipe: Dict[str, Any]) -> Dict[str, Any]:
    kinematics_xml = recipe["HelicityKinematics"]
    initialstate_xml = kinematics_xml["InitialState"]["Particle"]
    initial_state_yml = list()
    for state_def_xml in initialstate_xml:
        state_def_yml = {
            "Particle": state_def_xml["Name"],
            "ID": int(state_def_xml["Id"]),
        }
        initial_state_yml.append(state_def_yml)
    kinematics_yaml = {
        "Type": "Helicity",
        "InitialState": _to_state_list(kinematics_xml, "InitialState"),
        "FinalState": _to_state_list(kinematics_xml, "FinalState"),
    }
    return kinematics_yaml


def to_parameter_list(recipe: Dict[str, Any]) -> List[Dict[str, Any]]:
    intensity_dict = recipe["Intensity"]
    xml_parameter_defs = _extract_parameter_definitions_from_intensity(
        intensity_dict
    )
    parameter_dict: Dict[str, Dict[str, Any]] = dict()
    for parameter_xml in xml_parameter_defs:
        name = parameter_xml["Name"]
        if name not in parameter_dict:
            parameter_yml = dict()
            parameter_yml["Type"] = parameter_xml["Type"]
            parameter_yml["Value"] = parameter_xml["Value"]
            if "Fix" in parameter_xml:
                is_fix = bool(parameter_xml["Fix"])
                if is_fix:
                    parameter_yml["Fix"] = is_fix
            parameter_dict[name] = parameter_yml
    parameter_list = [
        {"Name": name, **value} for name, value in parameter_dict.items()
    ]
    return parameter_list


def to_particle_dict(recipe: Dict[str, Any]) -> Dict[str, Any]:
    particle_list_xml = recipe["ParticleList"]["Particle"]
    particle_list_xml = sorted(particle_list_xml, key=lambda i: i["Name"])
    particle_list_yml = dict()
    for xml_particle in particle_list_xml:
        name = str(xml_particle["Name"])
        pid = int(xml_particle["Pid"])
        parameters = xml_particle["Parameter"]
        mass = _to_parameter(parameters)
        quantum_numbers = list(xml_particle["QuantumNumber"])
        decay_info = xml_particle.get("DecayInfo", None)

        qn_key_map: Dict[str, Tuple[str, Callable]] = {
            "Charge": ("Charge", _to_scalar),
            "Spin": ("Spin", _to_scalar),
            "IsoSpin": ("IsoSpin", _to_isospin),
            "Parity": ("Parity", _to_scalar),
            "Cparity": ("Cparity", _to_scalar),
            "Gparity": ("Gparity", _to_scalar),
            "Strangeness": ("Strangeness", _to_scalar),
            "Charm": ("Charm", _to_scalar),
            "Bottomness": ("Bottomness", _to_scalar),
            "Topness": ("Topness", _to_scalar),
            "ElectronLN": ("ElectronLN", _to_scalar),
            "MuonLN": ("MuonLN", _to_scalar),
            "TauLN": ("TauLN", _to_scalar),
            "BaryonNumber": ("BaryonNumber", _to_scalar),
        }
        yaml_qn_dict = {"Spin": 0, "Charge": 0}
        for quantum_number in quantum_numbers:
            qn_type = quantum_number["Type"]
            for xml_key, (yaml_key, converter) in qn_key_map.items():
                if qn_type == xml_key:
                    value = converter(quantum_number)
                    if isinstance(value, (float, int)) and value == 0:
                        continue
                    if isinstance(value, dict) and value["Value"] == 0:
                        continue
                    yaml_qn_dict[yaml_key] = value
        particle_yml: Dict[str, Any] = dict()
        particle_yml["PID"] = pid
        particle_yml["Mass"] = mass
        if decay_info:
            parameters = decay_info.get("Parameter", list())
            for parameter in parameters:
                if parameter["Type"] == "Width":
                    particle_yml["Width"] = _to_parameter(parameter)
        particle_yml["QuantumNumbers"] = yaml_qn_dict
        particle_list_yml[name] = particle_yml
    return particle_list_yml


def _extract_parameter_definitions_from_intensity(
    definition: dict,
) -> List[Dict[str, Any]]:
    if not isinstance(definition, dict):
        return list()
    search_results = gen_dict_extract("Parameter", definition)
    parameter_defs = list()
    for item in search_results:
        if isinstance(item, list):
            parameter_defs.extend(item)
        else:
            parameter_defs.append(item)
    return parameter_defs


def gen_dict_extract(
    search_term: str, dictionary: Dict[str, Any]
) -> Generator:
    """See https://stackoverflow.com/a/29652561"""
    for key, value in dictionary.items():
        if key == search_term:
            yield value
        if isinstance(value, dict):
            for result in gen_dict_extract(search_term, value):
                yield result
        elif isinstance(value, list):
            for d in value:
                for result in gen_dict_extract(search_term, d):
                    yield result


def _to_scalar(
    definition: Dict[str, str], key: str = "Value"
) -> Union[float, int]:
    value = _downgrade_float(float(definition[key]))
    return value


def _downgrade_float(value: float) -> Union[float, int]:
    if value.is_integer():
        return int(value)
    return value


def _to_parameter(
    definition: Dict[str, Any]
) -> Union[float, Dict[str, float]]:
    """Use for extracting Mass and Width keys."""
    value = float(definition["Value"])
    error = float(definition.get("Error", 0.0))
    if error == 0.0:
        return value
    return {"Value": value, "Error": error}


def _to_isospin(definition: Dict[str, Any]) -> Union[float, Dict[str, float]]:
    """Isospin is 'stable', so always needs a projection."""
    value = _to_scalar(definition, "Value")
    if value == 0:
        return value
    return {
        "Value": value,
        "Projection": _to_scalar(definition, "Projection"),
    }


def _to_state_list(
    definition: Dict[str, Any], key: str
) -> List[Dict[str, Union[str, int]]]:
    state_list_xml = definition[key]["Particle"]
    state_list_yml = list()
    for state_def_xml in state_list_xml:
        state_def_yml = {
            "Particle": state_def_xml["Name"],
            "ID": int(state_def_xml["Id"]),
        }
        state_list_yml.append(state_def_yml)
    return state_list_yml


def _extract_intensity_component(definition: Dict[str, Any]) -> Dict[str, Any]:
    output_dict = dict()
    class_name = definition["Class"]
    if class_name == "StrengthIntensity":
        output_dict = _extract_intensity_component(definition["Intensity"])
    elif class_name == "NormalizedIntensity":
        output_dict["Class"] = class_name
        output_dict["Intensity"] = _extract_intensity_component(
            definition["Intensity"]
        )
    elif class_name == "IncoherentIntensity":
        output_dict["Class"] = class_name
        intensities_yml = list()
        for intensity in definition["Intensity"]:
            intensities_yml.append(_extract_intensity_component(intensity))
        output_dict["Intensities"] = intensities_yml
    elif class_name == "CoherentIntensity":
        output_dict["Class"] = class_name
        output_dict["Component"] = definition["Component"]
        amplitudes_xml = _safe_wrap_in_list(definition["Amplitude"])
        amplitudes_yml = [
            _extract_intensity_component(amplitude)
            for amplitude in amplitudes_xml
        ]
        output_dict["Amplitudes"] = amplitudes_yml
    elif class_name == "CoefficientAmplitude":
        output_dict["Class"] = class_name
        output_dict["Component"] = definition["Component"]
        parameters_xml = _safe_wrap_in_list(definition["Parameter"])
        parameter_names = [par["Name"] for par in parameters_xml]
        parameters_yml = dict()
        for name in parameter_names:
            type_name = name.split("_")[0]
            parameters_yml[type_name] = name
        output_dict["Parameters"] = parameters_yml
        amplitudes_xml = definition["Amplitude"]
        amplitudes_yml = _extract_intensity_component(amplitudes_xml)  # type: ignore
        output_dict["Amplitude"] = amplitudes_yml
    elif class_name == "SequentialAmplitude":
        output_dict["Class"] = class_name
        amplitudes_xml = _safe_wrap_in_list(definition["Amplitude"])
        amplitudes_yml = [
            _extract_intensity_component(amplitude)
            for amplitude in amplitudes_xml
        ]
        output_dict["Amplitudes"] = amplitudes_yml
    elif class_name == "HelicityDecay":
        output_dict["Class"] = class_name
        decay_particle = definition["DecayParticle"]
        decay_particle["Helicity"] = float(decay_particle["Helicity"])
        output_dict["DecayParticle"] = decay_particle
        decay_products = definition["DecayProducts"]["Particle"]
        decay_products = _safe_wrap_in_list(decay_products)
        for decay_product in decay_products:
            final_states = decay_product["FinalState"].split(" ")
            final_states = [int(state_id) for state_id in final_states]
            decay_product["FinalState"] = final_states
            decay_product["Helicity"] = float(decay_product["Helicity"])
        output_dict["DecayProducts"] = decay_products
        if "RecoilSystem" in definition:
            recoil_system = definition["RecoilSystem"]
            recoil_system["RecoilFinalState"] = int(
                recoil_system["RecoilFinalState"]
            )
            output_dict["RecoilSystem"] = recoil_system
    return output_dict


def _safe_wrap_in_list(input_object: Any) -> List[Any]:
    if isinstance(input_object, list):
        return input_object
    return [input_object]



from abc import ABC, abstractmethod
from functools import reduce
from copy import deepcopy

from numpy import arange
import logging

from .particle import (
    StateQuantumNumberNames,
    InteractionQuantumNumberNames,
    ParticlePropertyNames,
    ParticleDecayPropertyNames,
    QNNameClassMapping,
    QuantumNumberClasses,
    is_boson,
    Spin,
)


""" Functors for quantum number condition checks """


class AbstractConditionFunctor(ABC):
    @abstractmethod
    def check(self, qn_names, in_edges, out_edges, int_node):
        pass


class DefinedForAllEdges(AbstractConditionFunctor):
    def check(self, qn_names, in_edges, out_edges, int_node):
        for qn_name in qn_names:
            for edge in in_edges + out_edges:
                if qn_name not in edge:
                    return False
        return True


class DefinedForAllOutgoingEdges(AbstractConditionFunctor):
    def check(self, qn_names, in_edges, out_edges, int_node):
        for qn_name in qn_names:
            for edge in out_edges:
                if qn_name not in edge:
                    return False
        return True


class DefinedForInteractionNode(AbstractConditionFunctor):
    def check(self, qn_names, in_edges, out_edges, int_node):
        for qn_name in qn_names:
            if qn_name not in int_node:
                return False
        return True


class DefinedIfOtherQnNotDefinedInOutSeparate(AbstractConditionFunctor):
    """
    Implements logic for...
    """

    def __init__(self, other_qn_names):
        self.other_qn_names = other_qn_names

    def check(self, qn_names, in_edges, out_edges, int_node):
        return self.check_edge_set(
            qn_names, in_edges, int_node
        ) and self.check_edge_set(qn_names, out_edges, int_node)

    def check_edge_set(self, qn_names, edges, int_node):
        found_for_all = True
        for qn_name in self.other_qn_names:
            if isinstance(
                qn_name, (StateQuantumNumberNames, ParticlePropertyNames)
            ):
                found_for_all = True
                for edge_props in edges:
                    if not self.find_in_dict(qn_name, edge_props):
                        found_for_all = False
                        break
                if not found_for_all:
                    break
            else:
                if not self.find_in_dict(qn_name, int_node):
                    found_for_all = False
                    break

        if not found_for_all:
            for qn_name in qn_names:
                if isinstance(
                    qn_name, (StateQuantumNumberNames, ParticlePropertyNames)
                ):
                    for edge_props in edges:
                        if not self.find_in_dict(qn_name, edge_props):
                            return False
                else:
                    if not self.find_in_dict(qn_name, int_node):
                        return False
        return True

    def find_in_dict(self, name, props):
        found = False
        for key, val in props.items():
            if name == key and val is not None:
                found = True
                break
        return found


def is_particle_antiparticle_pair(pid1, pid2):
    # we just check if the pid is opposite in sign
    # this is a requirement of the pid numbers of course
    return pid1 == -pid2


class AbstractRule(ABC):
    def __init__(self, rule_name):
        self.rule_name = str(rule_name)
        self.required_qn_names = []
        self.qn_conditions = []
        self.specify_required_qns()

    def __repr__(self):
        return str(self.rule_name)

    def __str__(self):
        return str(self.rule_name)

    def get_qn_conditions(self):
        return self.qn_conditions

    @abstractmethod
    def specify_required_qns(self):
        pass

    def add_required_qn(self, qn_name, qn_condition_functions=[]):
        if not (
            isinstance(qn_name, StateQuantumNumberNames)
            or isinstance(qn_name, InteractionQuantumNumberNames)
            or isinstance(qn_name, ParticlePropertyNames)
            or isinstance(qn_name, ParticleDecayPropertyNames)
        ):
            raise TypeError(
                "qn_name has to be of type "
                + "ParticleQuantumNumberNames or "
                + "InteractionQuantumNumberNames or "
                + "ParticlePropertyNames"
            )
        self.required_qn_names.append(qn_name)
        for cond in qn_condition_functions:
            self.qn_conditions.append(([qn_name], cond))

    def get_required_qn_names(self):
        return self.required_qn_names

    @abstractmethod
    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        pass

    def check_requirements(self, in_edges, out_edges, int_node):
        for (qn_name_list, cond_functor) in self.get_qn_conditions():
            # part_props = [x for x in qn_name_list if isinstance(
            #    x, ParticlePropertyNames)]
            # if part_props:
            #    return False

            if not cond_functor.check(
                qn_name_list, in_edges, out_edges, int_node
            ):
                logging.debug(
                    "condition "
                    + str(cond_functor.__class__)
                    + " for quantum numbers "
                    + str(qn_name_list)
                    + " for rule "
                    + str(self.__class__)
                    + " not satisfied"
                )
                return False
        return True


class AdditiveQuantumNumberConservation(AbstractRule):
    """
    checks for the conservation of an additive quantum number such as electric
    charge, baryon number, lepton number

    :math:`\\sum q_{in} = \\sum q_{out}`
    """

    def __init__(self, qn_name):
        self.qn_name = qn_name
        super().__init__(qn_name.name + "Conservation")

    def specify_required_qns(self):
        self.add_required_qn(self.qn_name, [DefinedForAllEdges()])

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        in_qn_sum = sum(
            [
                part[self.qn_name]
                for part in ingoing_part_qns
                if self.qn_name in part
            ]
        )
        out_qn_sum = sum(
            [
                part[self.qn_name]
                for part in outgoing_part_qns
                if (self.qn_name in part)
            ]
        )
        return in_qn_sum == out_qn_sum


class ParityConservation(AbstractRule):
    def __init__(self):
        super().__init__("ParityConservation")

    def specify_required_qns(self):
        self.add_required_qn(
            StateQuantumNumberNames.Parity, [DefinedForAllEdges()]
        )
        self.add_required_qn(
            InteractionQuantumNumberNames.L, [DefinedForInteractionNode()]
        )

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        """ implements :math:`P_{in} = P_{out} \\cdot (-1)^L` """
        # is this valid for two outgoing particles only?
        parity_label = StateQuantumNumberNames.Parity
        parity_in = reduce(
            lambda x, y: x * y[parity_label], ingoing_part_qns, 1
        )
        parity_out = reduce(
            lambda x, y: x * y[parity_label], outgoing_part_qns, 1
        )
        ang_mom = interaction_qns[InteractionQuantumNumberNames.L].magnitude()
        if parity_in == (parity_out * (-1) ** ang_mom):
            return True
        return False


class ParityConservationHelicity(AbstractRule):
    def __init__(self):
        super().__init__("ParityConservationHelicity")

    def specify_required_qns(self):
        self.add_required_qn(
            StateQuantumNumberNames.Parity, [DefinedForAllEdges()]
        )
        self.add_required_qn(
            StateQuantumNumberNames.Spin, [DefinedForAllEdges()]
        )
        self.add_required_qn(
            InteractionQuantumNumberNames.ParityPrefactor,
            [DefinedForInteractionNode()],
        )

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        """
        Implements the check

        :math:`A_{-\\lambda_1-\\lambda_2} = P_1 P_2 P_3 (-1)^{S_2+S_3-S_1} A_{\\lambda_1\\lambda_2}`

        Notice that only the special case :math:`\\lambda_1=\\lambda_2=0`
        may return False
        """
        if len(ingoing_part_qns) == 1 and len(outgoing_part_qns) == 2:
            spin_label = StateQuantumNumberNames.Spin
            parity_label = StateQuantumNumberNames.Parity

            spins = [
                x[spin_label].magnitude()
                for x in ingoing_part_qns + outgoing_part_qns
            ]
            parity_product = reduce(
                lambda x, y: x * y[parity_label],
                ingoing_part_qns + outgoing_part_qns,
                1,
            )

            prefactor = parity_product * (-1.0) ** (
                spins[1] + spins[2] - spins[0]
            )

            daughter_hel = [
                0
                for x in outgoing_part_qns
                if x[spin_label].projection() == 0.0
            ]
            if len(daughter_hel) == 2:
                if prefactor == -1:
                    return False

            pf_label = InteractionQuantumNumberNames.ParityPrefactor
            return prefactor == interaction_qns[pf_label]
        return True


class CParityConservation(AbstractRule):
    def __init__(self):
        super().__init__("CParityConservation")

    def specify_required_qns(self):
        self.add_required_qn(StateQuantumNumberNames.Cparity)
        # the spin quantum number is required to check if the daughter
        # particles are fermions or bosons
        self.add_required_qn(
            StateQuantumNumberNames.Spin,
            [
                DefinedIfOtherQnNotDefinedInOutSeparate(
                    [StateQuantumNumberNames.Cparity]
                )
            ],
        )
        self.add_required_qn(
            InteractionQuantumNumberNames.L,
            [
                DefinedIfOtherQnNotDefinedInOutSeparate(
                    [StateQuantumNumberNames.Cparity]
                )
            ],
        )
        self.add_required_qn(
            InteractionQuantumNumberNames.S,
            [
                DefinedIfOtherQnNotDefinedInOutSeparate(
                    [StateQuantumNumberNames.Cparity]
                )
            ],
        )
        self.add_required_qn(
            ParticlePropertyNames.Pid,
            [
                DefinedIfOtherQnNotDefinedInOutSeparate(
                    [StateQuantumNumberNames.Cparity]
                )
            ],
        )

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        """ implements :math:`C_{in} = C_{out}` """
        cparity_in = self.get_cparity_multiparticle(
            ingoing_part_qns, interaction_qns
        )
        if cparity_in is None:
            return True

        cparity_out = self.get_cparity_multiparticle(
            outgoing_part_qns, interaction_qns
        )
        if cparity_out is None:
            return True

        return cparity_in == cparity_out

    def get_cparity_multiparticle(self, part_qns, interaction_qns):
        cparity_label = StateQuantumNumberNames.Cparity
        pid_label = ParticlePropertyNames.Pid
        ang_mom_label = InteractionQuantumNumberNames.L
        int_spin_label = InteractionQuantumNumberNames.S

        no_cpar_part = [
            part_qns.index(x)
            for x in part_qns
            if cparity_label not in x or x[cparity_label] is None
        ]
        # if all states have c parity defined, then just multiply them
        if not no_cpar_part:
            return reduce(lambda x, y: x * y[cparity_label], part_qns, 1)

        # two particle case
        if len(part_qns) == 2:
            if is_particle_antiparticle_pair(
                part_qns[0][pid_label], part_qns[1][pid_label]
            ):
                ang_mom = interaction_qns[ang_mom_label].magnitude()
                # if boson
                if is_boson(part_qns[0]):
                    return (-1) ** ang_mom
                else:
                    coupled_spin = interaction_qns[int_spin_label].magnitude()
                    return (-1) ** (ang_mom + coupled_spin)
        """elif len(no_cpar_part) > 0 and len(no_cpar_part) % 2 == 0:
            # does this also work for more than 2 particles?
            # try to find pairs of particle antiparticle
            pids = [x[pid_label] for x in part_qns]
            while pids:
                found = False
                ref_pid = pids.pop()
                for x in pids:
                    if is_particle_antiparticle_pair(ref_pid, x):
                        found = True
                        break
                if not found:
                    break"""

        return None


class GParityConservation(AbstractRule):
    def __init__(self):
        super().__init__("GParityConservation")

    def specify_required_qns(self):
        self.add_required_qn(StateQuantumNumberNames.Gparity)
        # the spin quantum number is required to check if the daughter
        # particles are fermions or bosons
        self.add_required_qn(
            StateQuantumNumberNames.Spin,
            [
                DefinedIfOtherQnNotDefinedInOutSeparate(
                    [StateQuantumNumberNames.Gparity]
                )
            ],
        )
        self.add_required_qn(
            InteractionQuantumNumberNames.L,
            [
                DefinedIfOtherQnNotDefinedInOutSeparate(
                    [StateQuantumNumberNames.Gparity]
                )
            ],
        )
        self.add_required_qn(
            InteractionQuantumNumberNames.S,
            [
                DefinedIfOtherQnNotDefinedInOutSeparate(
                    [StateQuantumNumberNames.Gparity]
                )
            ],
        )
        self.add_required_qn(
            StateQuantumNumberNames.IsoSpin,
            [
                DefinedIfOtherQnNotDefinedInOutSeparate(
                    [StateQuantumNumberNames.Gparity]
                )
            ],
        )
        self.add_required_qn(
            ParticlePropertyNames.Pid,
            [
                DefinedIfOtherQnNotDefinedInOutSeparate(
                    [StateQuantumNumberNames.Gparity]
                )
            ],
        )

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        """ implements :math:`G_{in} = G_{out}` """
        gparity_label = StateQuantumNumberNames.Gparity
        no_gpar_inpart = [
            ingoing_part_qns.index(x)
            for x in ingoing_part_qns
            if gparity_label not in x or x[gparity_label] is None
        ]
        no_gpar_outpart = [
            outgoing_part_qns.index(x)
            for x in outgoing_part_qns
            if gparity_label not in x or x[gparity_label] is None
        ]
        # if all states have g parity defined, then just multiply them
        if not no_gpar_inpart + no_gpar_outpart:
            in_gpar = reduce(
                lambda x, y: x * y[gparity_label], ingoing_part_qns, 1
            )
            out_gpar = reduce(
                lambda x, y: x * y[gparity_label], outgoing_part_qns, 1
            )
            return in_gpar == out_gpar

        # two particle case
        particle_counts = (len(ingoing_part_qns), len(outgoing_part_qns))
        if particle_counts == (1, 2):
            if gparity_label in ingoing_part_qns[0]:
                out_gpar = self.check_multistate_gparity(
                    ingoing_part_qns, outgoing_part_qns, interaction_qns
                )
                in_gpar = ingoing_part_qns[0][gparity_label]
                if out_gpar is not None and in_gpar is not None:
                    return out_gpar == in_gpar

        if particle_counts == (2, 1):
            if gparity_label in outgoing_part_qns[0]:
                in_gpar = self.check_multistate_gparity(
                    outgoing_part_qns, ingoing_part_qns, interaction_qns
                )
                out_gpar = outgoing_part_qns[0][gparity_label]
                if out_gpar is not None and in_gpar is not None:
                    return out_gpar == in_gpar
        return True

    def check_multistate_gparity(
        self, single_state_qns, double_state_qns, interaction_qns
    ):
        isospin_label = StateQuantumNumberNames.IsoSpin
        pid_label = ParticlePropertyNames.Pid
        ang_mom_label = InteractionQuantumNumberNames.L
        int_spin_label = InteractionQuantumNumberNames.S
        if is_particle_antiparticle_pair(
            double_state_qns[0][pid_label], double_state_qns[1][pid_label]
        ):
            ang_mom = interaction_qns[ang_mom_label].magnitude()
            isospin = single_state_qns[0][isospin_label].magnitude()
            # if boson
            if is_boson(double_state_qns[0]):
                return (-1) ** (ang_mom + isospin)
            else:
                coupled_spin = interaction_qns[int_spin_label].magnitude()
                return (-1) ** (ang_mom + coupled_spin + isospin)
        return None


class IdenticalParticleSymmetrization(AbstractRule):
    def __init__(self):
        super().__init__("IdenticalParticleSymmetrization")

    def specify_required_qns(self):
        self.add_required_qn(StateQuantumNumberNames.Parity)
        self.add_required_qn(
            ParticlePropertyNames.Pid, [DefinedForAllOutgoingEdges()]
        )
        self.add_required_qn(
            StateQuantumNumberNames.Spin, [DefinedForAllOutgoingEdges()]
        )

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        if self.check_particles_identical(outgoing_part_qns):
            parity_label = StateQuantumNumberNames.Parity

            if is_boson(outgoing_part_qns[0]):
                # we have a boson, check if parity of mother is even
                parity = ingoing_part_qns[0][parity_label]
                if parity == -1:
                    # if its odd then return False
                    return False
            else:
                # its fermion
                parity = ingoing_part_qns[0][parity_label]
                if parity == 1:
                    return False

        return True

    def check_particles_identical(self, particles):
        # check if pids and spins match
        pid_label = ParticlePropertyNames.Pid
        spin_label = StateQuantumNumberNames.Spin
        reference_pid = particles[0][pid_label]
        reference_spin = particles[0][spin_label]
        for p in particles[1:]:
            if p[pid_label] != reference_pid:
                return False
            if p[spin_label] != reference_spin:
                return False
        return True


class SpinConservation(AbstractRule):
    """
    Implements conservation of a spin-like quantum number for a two body decay
    (coupling of two particle states). See :py:meth:`check` for details.
    """

    def __init__(self, spinlike_qn, use_projection=True):
        if not isinstance(spinlike_qn, StateQuantumNumberNames):
            raise TypeError(
                "Expecting Emum of the type \
                ParticleQuantumNumberNames for spinlike_qn"
            )
        if spinlike_qn not in QNNameClassMapping:
            raise ValueError("spinlike_qn is not associated with a QN class")
        if QNNameClassMapping[spinlike_qn] is not QuantumNumberClasses.Spin:
            raise ValueError("spinlike_qn is not of class Spin")
        self.spinlike_qn = spinlike_qn
        self.use_projection = use_projection
        super().__init__(spinlike_qn.name + "Conservation")

    def specify_required_qns(self):
        self.add_required_qn(self.spinlike_qn, [DefinedForAllEdges()])
        # for actual spins we include the angular momentum
        if self.spinlike_qn is StateQuantumNumberNames.Spin:
            self.add_required_qn(
                InteractionQuantumNumberNames.L, [DefinedForInteractionNode()]
            )
            self.add_required_qn(
                InteractionQuantumNumberNames.S, [DefinedForInteractionNode()]
            )

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        """
        implements :math:`|S_1 - S_2| \\leq S \\leq |S_1 + S_2|` and optionally
        :math:`|L - S| \\leq J \\leq |L + S|`. Also checks
        :math:`M_1 + M_2 = M` and if Clebsch-Gordan coefficients are 0
        """
        spin_label = self.spinlike_qn

        in_spins = [x[spin_label] for x in ingoing_part_qns]
        out_spins = [x[spin_label] for x in outgoing_part_qns]
        if self.use_projection and not self.check_projections(
            in_spins, out_spins
        ):
            return False
        return self.check_magnitude(in_spins, out_spins, interaction_qns)

    def check_projections(self, in_part, out_part):
        in_proj = [x.projection() for x in in_part]
        out_proj = [x.projection() for x in out_part]
        return sum(in_proj) == sum(out_proj)

    def check_magnitude(self, in_part, out_part, interaction_qns):
        in_tot_spins = self.calculate_total_spins(in_part, interaction_qns)
        out_tot_spins = self.calculate_total_spins(out_part, interaction_qns)
        matching_spins = in_tot_spins.intersection(out_tot_spins)
        if len(matching_spins) > 0:
            return True
        return False

    def calculate_total_spins(self, part_list, interaction_qns):
        total_spins = set()
        if len(part_list) == 1:
            if self.use_projection:
                total_spins.add(part_list[0])
            else:
                total_spins.add(Spin(part_list[0].magnitude(), 0))
        else:
            # first couple all spins together
            spins_daughters_coupled = set()
            spin_list = deepcopy(part_list)
            while spin_list:
                if spins_daughters_coupled:
                    temp_coupled_spins = set()
                    tempspin = spin_list.pop()
                    for s in spins_daughters_coupled:
                        coupled_spins = self.spin_couplings(s, tempspin)
                        temp_coupled_spins.update(coupled_spins)
                    spins_daughters_coupled = temp_coupled_spins
                else:
                    spins_daughters_coupled.add(spin_list.pop())
            if InteractionQuantumNumberNames.L in interaction_qns:
                L = interaction_qns[InteractionQuantumNumberNames.L]
                S = interaction_qns[InteractionQuantumNumberNames.S]
                if self.use_projection:
                    if S in spins_daughters_coupled:
                        total_spins.update(self.spin_couplings(S, L))
                else:
                    if S.magnitude() in [
                        x.magnitude() for x in spins_daughters_coupled
                    ]:
                        total_spins.update(self.spin_couplings(S, L))
            else:
                if self.use_projection:
                    total_spins = spins_daughters_coupled
                else:
                    total_spins = [
                        Spin(x.magnitude(), 0.0)
                        for x in spins_daughters_coupled
                    ]
        return total_spins

    def spin_couplings(self, spin1, spin2):
        """
        implements the coupling of two spins
        :math:`|S_1 - S_2| \\leq S \\leq |S_1 + S_2|` and :math:`M_1 + M_2 = M`
        """
        j1 = spin1.magnitude()
        j2 = spin2.magnitude()
        if self.use_projection:
            m = spin1.projection() + spin2.projection()
            possible_spins = [
                Spin(x, m)
                for x in arange(abs(j1 - j2), j1 + j2 + 1, 1).tolist()
                if x >= abs(m)
            ]

            return [
                x
                for x in possible_spins
                if not is_clebsch_gordan_coefficient_zero(spin1, spin2, x)
            ]
        else:
            return [
                Spin(x, 0)
                for x in arange(abs(j1 - j2), j1 + j2 + 1, 1).tolist()
            ]


def is_clebsch_gordan_coefficient_zero(spin1, spin2, spin_coupled):
    m1 = spin1.projection()
    j1 = spin1.magnitude()
    m2 = spin2.projection()
    j2 = spin2.magnitude()
    m = spin_coupled.projection()
    j = spin_coupled.magnitude()
    iszero = False
    if (j1 == j2 and m1 == m2) or (m1 == 0.0 and m2 == 0.0):
        if abs(j - j1 - j2) % 2 == 1:
            iszero = True
    elif j1 == j and m1 == -m:
        if abs(j2 - j1 - j) % 2 == 1:
            iszero = True
    elif j2 == j and m2 == -m:
        if abs(j1 - j2 - j) % 2 == 1:
            iszero = True
    return iszero


class ClebschGordanCheckHelicityToCanonical(AbstractRule):
    """
    implements clebsch gordan checks for :math:`S_1, S_2` to :math:`S` and the
    :math:`L,S` to :math:`J` coupling based on the conversion of helicity to
    canonical amplitude sums
    """

    def __init__(self):
        super().__init__("ClebschGordanCheckHelicityToCanonical")

    def specify_required_qns(self):
        self.add_required_qn(
            StateQuantumNumberNames.Spin, [DefinedForAllEdges()]
        )
        self.add_required_qn(
            InteractionQuantumNumberNames.L, [DefinedForInteractionNode()]
        )
        self.add_required_qn(
            InteractionQuantumNumberNames.S, [DefinedForInteractionNode()]
        )

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        if len(ingoing_part_qns) == 1 and len(outgoing_part_qns) == 2:
            spin_label = StateQuantumNumberNames.Spin
            in_spins = [x[spin_label] for x in ingoing_part_qns]
            out_spins = [x[spin_label] for x in outgoing_part_qns]
            out_spins[1] = Spin(
                out_spins[1].magnitude(), -out_spins[1].projection()
            )
            helicity_diff = sum([x.projection() for x in out_spins])
            L = interaction_qns[InteractionQuantumNumberNames.L]
            S = interaction_qns[InteractionQuantumNumberNames.S]
            if S.magnitude() < abs(helicity_diff) or in_spins[
                0
            ].magnitude() < abs(helicity_diff):
                return False
            S = Spin(S.magnitude(), helicity_diff)
            if is_clebsch_gordan_coefficient_zero(
                out_spins[0], out_spins[1], S
            ):
                return False
            in_spins[0] = Spin(in_spins[0].magnitude(), helicity_diff)
            return not is_clebsch_gordan_coefficient_zero(L, S, in_spins[0])
        return False


class HelicityConservation(AbstractRule):
    def __init__(self):
        super().__init__("HelicityConservation")

    def specify_required_qns(self):
        self.add_required_qn(
            StateQuantumNumberNames.Spin, [DefinedForAllEdges()]
        )

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        """
        implements :math:`|\\lambda_2-\\lambda_3| \\leq S_1`
        """
        if len(ingoing_part_qns) == 1 and len(outgoing_part_qns) == 2:
            spin_label = StateQuantumNumberNames.Spin

            mother_spin = ingoing_part_qns[0][spin_label].magnitude()
            daughter_hel = [
                x[spin_label].projection() for x in outgoing_part_qns
            ]
            if mother_spin >= abs(daughter_hel[0] - daughter_hel[1]):
                return True
        return False


class GellMannNishijimaRule(AbstractRule):
    def __init__(self):
        super().__init__("GellMannNishijimaRule")

    def specify_required_qns(self):
        self.add_required_qn(
            StateQuantumNumberNames.Charge, [DefinedForAllEdges()]
        )
        self.add_required_qn(
            StateQuantumNumberNames.IsoSpin, [DefinedForAllEdges()]
        )
        self.add_required_qn(StateQuantumNumberNames.Strangeness)
        self.add_required_qn(StateQuantumNumberNames.Charm)
        self.add_required_qn(StateQuantumNumberNames.Bottomness)
        self.add_required_qn(StateQuantumNumberNames.Topness)
        self.add_required_qn(StateQuantumNumberNames.BaryonNumber)
        self.add_required_qn(StateQuantumNumberNames.ElectronLN)
        self.add_required_qn(StateQuantumNumberNames.MuonLN)
        self.add_required_qn(StateQuantumNumberNames.TauLN)

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        """
        checks the Gell-Mann–Nishijima formula :math:`Q=I_3+\\frac{Y}{2}` for
        each particle.
        """
        charge_label = StateQuantumNumberNames.Charge
        isospin_label = StateQuantumNumberNames.IsoSpin

        eln = StateQuantumNumberNames.ElectronLN
        mln = StateQuantumNumberNames.MuonLN
        tln = StateQuantumNumberNames.TauLN

        for particle in ingoing_part_qns + outgoing_part_qns:
            if (
                sum(
                    [
                        abs(particle[x])
                        for x in [eln, mln, tln]
                        if x in particle
                    ]
                )
                > 0.0
            ):
                # if particle is a lepton then skip the check
                continue
            isospin_3 = 0
            if isospin_label in particle:
                isospin_3 = particle[isospin_label].projection()
            if float(particle[charge_label]) != (
                isospin_3 + 0.5 * self.calculate_hypercharge(particle)
            ):
                return False
        return True

    def calculate_hypercharge(self, particle):
        """
        calculates the hypercharge :math:`Y=S+C+B+T+B`
        """
        qn_labels = [
            StateQuantumNumberNames.Strangeness,
            StateQuantumNumberNames.Charm,
            StateQuantumNumberNames.Bottomness,
            StateQuantumNumberNames.Topness,
            StateQuantumNumberNames.BaryonNumber,
        ]
        qn_values = [particle[x] for x in qn_labels if x in particle]
        return sum(qn_values)


class MassConservation(AbstractRule):
    def __init__(self, width_factor=3):
        self.width_factor = width_factor
        super().__init__("MassConservation")

    def specify_required_qns(self):
        self.add_required_qn(
            ParticlePropertyNames.Mass, [DefinedForAllEdges()]
        )
        self.add_required_qn(ParticleDecayPropertyNames.Width)

    def check(self, ingoing_part_qns, outgoing_part_qns, interaction_qns):
        """
        implements the mass check

        :math:`M_{out} - N \\cdot W_{out} < M_{in} + N \\cdot W_{in}`

        It makes sure that the net mass outgoing state :math:`M_{out}` is
        smaller than the net mass of the ingoing state :math:`M_{in}`. Also
        the width :math:`W` of the states is taken into account.
        """
        mass_label = ParticlePropertyNames.Mass
        width_label = ParticleDecayPropertyNames.Width

        mass_in = sum([x[mass_label] for x in ingoing_part_qns])
        width_in = sum(
            [x[width_label] for x in ingoing_part_qns if width_label in x]
        )
        mass_out = sum([x[mass_label] for x in outgoing_part_qns])
        width_out = sum(
            [x[width_label] for x in outgoing_part_qns if width_label in x]
        )

        return (mass_out - self.width_factor * width_out) < (
            mass_in + self.width_factor * width_in
        )



import json
import logging
from os.path import dirname, realpath

import pytest

import yaml

from expertsystem.amplitude.helicitydecay import HelicityAmplitudeGenerator
from expertsystem.ui.system_control import (
    InteractionTypes,
    StateTransitionManager,
)

logging.basicConfig(level=logging.ERROR)

SCRIPT_PATH = dirname(realpath(__file__))


def create_amplitude_generator():
    initial_state = [("J/psi", [-1, 1])]
    final_state = [("gamma", [-1, 1]), ("pi0", [0]), ("pi0", [0])]

    tbd_manager = StateTransitionManager(initial_state, final_state, ["f"])
    tbd_manager.set_allowed_interaction_types(
        [InteractionTypes.Strong, InteractionTypes.EM]
    )
    graph_interaction_settings_groups = tbd_manager.prepare_graphs()
    solutions, _ = tbd_manager.find_solutions(
        graph_interaction_settings_groups
    )

    amplitude_generator = HelicityAmplitudeGenerator()
    amplitude_generator.generate(solutions)
    return amplitude_generator


def write_load_yaml() -> dict:
    amplitude_generator = create_amplitude_generator()
    output_filename = "JPsiToGammaPi0Pi0.yml"
    amplitude_generator.write_to_file(output_filename)
    with open(output_filename, "rb") as input_file:
        imported_dict = yaml.load(input_file, Loader=yaml.FullLoader)
    return imported_dict


def load_expected_recipe() -> dict:
    expected_recipe_file = f"{SCRIPT_PATH}/expected_recipe.yml"
    with open(expected_recipe_file, "rb") as input_file:
        expected_dict = yaml.load(input_file, Loader=yaml.FullLoader)
    return expected_dict


def equalize_dict(input_dict):
    output_dict = json.loads(json.dumps(input_dict, sort_keys=True))
    return output_dict


class TestHelicityAmplitudeGeneratorYAML:
    amplitude_generator = create_amplitude_generator()
    imported_dict = write_load_yaml()
    expected_dict = load_expected_recipe()

    def test_not_implemented_writer(self):
        with pytest.raises(NotImplementedError):
            self.amplitude_generator.write_to_file("JPsiToGammaPi0Pi0.csv")

    def test_create_recipe_dict(self):
        recipe = self.amplitude_generator._create_recipe_dict()
        assert len(recipe) == 3

    def test_particle_section(self):
        particle_list = self.imported_dict["ParticleList"]
        gamma = particle_list["gamma"]
        assert gamma["PID"] == 22
        assert gamma["Mass"] == 0.0
        gamma_qns = gamma["QuantumNumbers"]
        assert gamma_qns["Spin"] == 1
        assert gamma_qns["Charge"] == 0
        assert gamma_qns["Parity"] == -1
        assert gamma_qns["Cparity"] == -1

        f0_980 = particle_list["f0(980)"]
        assert f0_980["Width"] == 0.07

        pi0_qns = particle_list["pi0"]["QuantumNumbers"]
        assert pi0_qns["IsoSpin"]["Value"] == 1
        assert pi0_qns["IsoSpin"]["Projection"] == 0

    def test_kinematics_section(self):
        kinematics = self.imported_dict["Kinematics"]
        initial_state = kinematics["InitialState"]
        final_state = kinematics["FinalState"]
        assert kinematics["Type"] == "Helicity"
        assert len(initial_state) == 1
        assert initial_state[0]["Particle"] == "J/psi"
        assert len(final_state) == 3

    def test_parameter_section(self):
        parameter_list = self.imported_dict["Parameters"]
        assert len(parameter_list) == 17
        for parameter in parameter_list:
            assert "Name" in parameter
            assert "Value" in parameter

    def test_dynamics_section(self):
        dynamics = self.imported_dict["Dynamics"]
        assert len(dynamics) == 5

        jpsi = dynamics["J/psi"]
        assert jpsi["Type"] == "NonDynamic"
        assert jpsi["FormFactor"]["Type"] == "BlattWeisskopf"
        assert jpsi["FormFactor"]["MesonRadius"] == 1.0

        f0_980 = dynamics["f0(980)"]
        assert f0_980["Type"] == "RelativisticBreitWigner"
        assert f0_980["FormFactor"]["Type"] == "BlattWeisskopf"
        assert f0_980["FormFactor"]["MesonRadius"] == {
            "Fix": True,
            "Max": 2.0,
            "Min": 0.0,
            "Value": 1.0,
        }

    def test_intensity_section(self):
        intensity = self.imported_dict["Intensity"]
        assert intensity["Class"] == "NormalizedIntensity"

        intensity = intensity["Intensity"]
        assert intensity["Class"] == "IncoherentIntensity"
        assert len(intensity["Intensities"]) == 4

    @pytest.mark.parametrize(
        "section", ["ParticleList", "Dynamics", "Kinematics"],
    )
    def test_expected_recipe_shape(self, section):
        expected_section = equalize_dict(self.expected_dict[section])
        imported_section = equalize_dict(self.imported_dict[section])
        if isinstance(expected_section, dict):
            assert expected_section.keys() == imported_section.keys()
            imported_items = list(imported_section.values())
            expected_items = list(expected_section.values())
        else:
            expected_items = list(expected_section)
            imported_items = list(imported_section)
        is_parameter_section = False
        if section == "Parameters":
            is_parameter_section = True
            imported_items = imported_items[1:]
        assert len(imported_items) == len(expected_items)
        for imported, expected in zip(imported_items, expected_items):
            if is_parameter_section:
                imported["Name"] = imported["Name"].replace("_-1", "_1")
            assert imported == expected



"""
The expert system facilitates building an amplitude model.

An amplitude model describes the reaction process you want to study with
partial wave analysis techniques. The responsibility of the expert system is to
give advice on the form of an amplitude model based on the problem set one
defines for a reaction process (initial state, final state, allowed
interactions, intermediate states, etc.). Internally, the system propagates the
quantum numbers through the reaction graph, while satisfying the specified
conservation rules.

Afterwards, the amplitude model of the expert system can be exported. This
amplitude model can then for instance be used to generate a data set (toy Monte
Carlo) for this specific reaction process, or to optimize ('fit') its
parameters so that they resemble the data set as good as possible.
"""


__all__ = [
    "amplitude",
    "state",
    "topology",
    "ui",
]


import sys

from . import amplitude
from . import state
from . import topology
from . import ui


def __check_python_version():
    def print_message_and_exit():
        print(
            "You are running python "
            + str(sys.version_info[0])
            + "."
            + str(sys.version_info[1])
            + "."
            + str(sys.version_info[2])
        )
        print("The expertsystem module requires Python 3.6 or higher!")
        sys.exit()

    if sys.version_info.major < 3:
        print_message_and_exit()
    elif sys.version_info.major == 3 and sys.version_info.minor < 6:
        print_message_and_exit()


if __name__ == "__main__":
    __check_python_version()



"""
A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
"""

import setuptools

DATA_FILES = [
    ("share", ["particle_list.xml", "particle_list.yml"]),
]

INSTALL_REQUIRES = [
    "numpy",
    "progress",
    "xmltodict",
]


def long_description():
    """Parse long description from readme."""
    with open("README.md", "r") as readme_file:
        return readme_file.read()


setuptools.setup(
    name="expertsystem",
    version="0.1.1-alpha",
    author="The ComPWA team",
    maintainer_email="compwa-admin@ep1.rub.de",
    url="https://github.com/ComPWA/expertsystem",
    long_description=long_description(),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    license="GPLv3 or later",
    python_requires=">=3.6",
    tests_require=["pytest"],
    install_requires=INSTALL_REQUIRES,
    data_files=DATA_FILES,
    package_data=dict(DATA_FILES),
    include_package_data=True,
)



from expertsystem.topology.graph import InteractionNode
from expertsystem.topology.topologybuilder import (
    SimpleStateTransitionTopologyBuilder,
)


class TestSimpleStateBuilder:  # pylint: disable=no-self-use
    def test_two_body_states(self):
        two_body_decay_node = InteractionNode("TwoBodyDecay", 1, 2)

        simple_builder = SimpleStateTransitionTopologyBuilder(
            [two_body_decay_node]
        )

        all_graphs = simple_builder.build_graphs(1, 3)

        assert len(all_graphs) == 1
